/****************************************************************************
 *
 * $Id:
 *
 * Copyright (C) 1998-2006 Lasmea. All rights reserved.
 *
 * This software was developed at:
 * Institut Pascal- UMR 6602
 * 63 175 Aubière Cedex
 * http://ip.univ-bpclermont.fr
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE included in the packaging of this file.
 *
 * Licensees holding valid IPDS Professional Edition licenses may
 * use this file in accordance with the IPDS Commercial License
 * Agreement provided with the Software.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Contact . if any conditions of this licensing are
 * not clear to you.
 *
 *
 * Author: Jonathan Courbon
 *
 *****************************************************************************/

#ifndef MAINWINDOW_H
#define MAINWINDOW_H


#include "config.h"
#if QTDETECTEDVERSION == 4
    #include <QMainWindow>
#else
    #include <QtWidgets/QMainWindow>
#endif



#include "qthreadexec.h"
#include "imageloader.h"
#include "containers.h"
#include "imagecombination.h"
#include "trajectorytoimage.h"
#include "lidarloader.h"

namespace Ui {
    class MainWindow;
}

class MainWindow : public QMainWindow {
    Q_OBJECT
public:


    /*!
     \brief

     \param folderpath path of the folder where data are saved
     \param akey name of the key
     \param aimwidth width of A SINGLE data viewer
     \param aimheight height of A SINGLE data viewer
     \param fps: framerate (in Hz). If fps==-1: maximum velocity
    */
    MainWindow(std::string folderpath,std::string akey, int aimwidth=320,int aimheight=240,int afps=-1,int typetimedisplay=0,ImageCombination::MODE mode=ImageCombination::ZERO,int astep=1,bool addSensorsName=true,double maxLaserDist=30,long long int startingtime=-1,long long int endingtime=-1,QWidget *parent = 0);
    ~MainWindow();

protected:
    void changeEvent(QEvent *e);

    // manage keyboard actions
    void keyPressEvent(QKeyEvent* event);

private slots:
    void updateData();
    void updateData(QImage im);

    void startUpdate();

signals:
    void terminateThreads();
    void loadImage_f_l_cam(QString);
    void loadImage_f_r_cam(QString);
    void loadImage_b_cam(QString);
    void loadImage_cata_cam(QString);
    void loadImage_fe_cam(QString);
    void loadImage_webcam(QString);
    void loadImage_f_m_cam(QString);

    void addTime(long long int);

    void loadImage_hz_laser(QString);
    void loadImage_incl_laser(QString);

    void saveImage(QString);

    void addNodeGPS(double positionX, double positionY,QColor color);
    void addNodeLowCostGPS(double positionX, double positionY,QColor color);
    void addNodeOdo(double positionX, double positionY,QColor color);

    void addAdditionnalImage(QImage im,bool keepImage);

private:

    ConfigTable ctable;
    void quit();

    void getImageFilename(Unit & unit,std::string prefix,std::string suffix,QString & filename);
    void getLidarFilename(Unit & unit,std::string prefix,QString & filename);

    void loadImage(QString filename,std::string ent);

    Ui::MainWindow *ui;

    void initialiseTraj(std::string entity,TrajectoryToImage& traj,std::vector<Unit> & data);

    void loadSatelliteImage(QString filename,int height,int width);

    int step;

    int fps;


    ImageLoader loader_f_l_cam;
    QThreadExec threadCamera_f_l_cam;

    ImageLoader loader_f_r_cam;
    QThreadExec threadCamera_f_r_cam;

    ImageLoader loader_f_m_cam;
    QThreadExec threadCamera_f_m_cam;

    ImageLoader loader_b_cam;
    QThreadExec threadCamera_b_cam;

    ImageLoader loader_cata_cam;
    QThreadExec threadCamera_cata_cam;

    ImageLoader loader_fe_cam;
    QThreadExec threadCamera_fe_cam;

    ImageLoader loader_webcam;
    QThreadExec threadCamera_webcam;

    ImageCombination imagecombin;
    QThreadExec threadImagecombin;

    TrajectoryToImage trajGPS;
    TrajectoryToImage trajLowCostGPS;
    TrajectoryToImage trajOdo;

    LidarLoader loaderLidHz;
    LidarLoader loaderLidIncl;

    std::vector<Unit> keyData;

    unsigned int currentkeyindex;
    std::string keyprefix,keysuffix;

    std::string key;


    double travDistPreviousNode;
    double threshDist;
    double threshTravelled;
    bool drawLoops;

    int currentNodeIndex;
    int idPreviousNode;

    QMutex mutPause_;
    bool pause_;
    bool stillstarted_;

    QImage currentImage_;
    long long int  currentImageTime_;

    bool singleSensor_;

    int typetimedisplay_;

};

#endif // MAINWINDOW_H
