/****************************************************************************
 *
 * $Id:
 *
 * Copyright (C) 1998-2006 Lasmea. All rights reserved.
 *
 * This software was developed at:
 * Institut Pascal- UMR 6602
 * 63 175 Aubière Cedex
 * http://ip.univ-bpclermont.fr
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE included in the packaging of this file.
 *
 * Licensees holding valid IPDS Professional Edition licenses may
 * use this file in accordance with the IPDS Commercial License
 * Agreement provided with the Software.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Contact . if any conditions of this licensing are
 * not clear to you.
 *
 *
 * Author: Jonathan Courbon - Hemanth Korrapati -
 *
 *****************************************************************************/


#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "config.h"
#include "containers.h"
#include "sensor_data.h"

#include <algorithm>
#include <cassert>
#include <boost/assign/std/vector.hpp>
#include <boost/assign/std/vector.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/bind.hpp>

#include <fstream>
#include <cstdlib>

#if QTDETECTEDVERSION == 4
    #include <QTimer>
    #include <QDir>
    #include <QKeyEvent>
#else
    #include <QtCore/QTimer>
    #include <QtCore/QDir>
    #include <QtGui/QKeyEvent>
#endif



#include <math.h>
#include <sstream>
#include <iomanip>

using namespace std;
using namespace boost;
using namespace boost::assign;

Q_DECLARE_METATYPE(long long int);


MainWindow::MainWindow(std::string folderpath,std::string akey,int aimwidth,int aimheight,int afps,int typetimedisplay,ImageCombination::MODE mode,int astep,bool addSensorsName,double maxLaserDist,long long int startingtime,long long int endingtime,QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{

    qRegisterMetaType<long long int >("long long int");
    /*
                INITIALISATION
     */
    // set up the interface
    ui->setupUi(this);

    ConfigTable ctable;
    std::stringstream msg;
    msg << CONFIGFOLDERPATH << "/config_table";
    ctable.readTable(msg.str());


    // copy elements
    key=akey;
    step=astep;
    fps=afps;

    // not yet started
    stillstarted_=false;

    // initial state: pause
    pause_=true;



    /*
                INITIALISATION OF THE IMAGE COMBINATION MODULE
     */

    // size returned by the module
    int returnedFullImageWidth,returnedFullImageHeight;

    // set mode
    imagecombin.setParameters(aimwidth,aimheight,returnedFullImageWidth,returnedFullImageHeight,mode);
    cout <<"Size of the displayed image in he MainWindow: " << returnedFullImageWidth << " x "  <<returnedFullImageHeight << endl;

    if(returnedFullImageHeight!=0)
    {
        // resize widget
        ui->widget_.setHeight(returnedFullImageHeight);
        this->resize(returnedFullImageWidth+20,returnedFullImageHeight+100);
        ui->gridLayoutWidget->setGeometry(QRect(0, 0,returnedFullImageWidth+20, returnedFullImageHeight+100));
    }

    // set save
    imagecombin.setSave(false);
    // set time display mode
    typetimedisplay_=typetimedisplay;
    imagecombin.setTypeTimeDisplay(typetimedisplay);
    // addition of the sensor name
    imagecombin.setAddSensorsNames(addSensorsName);

    /*
                FILE READING + SYNCHRONISATION
     */

    TimestampInterval interval(startingtime,endingtime);


    // extract readings for the key
    if(strcmp(key.c_str(),"time")!=0)
    {
        string timestamps_file_name=ctable.table[key].folderName+ctable.table[key].suffix;
        string sensor_data_folder=folderpath+"/"+ctable.table[key].folderName+"/";
        getReadings(ctable.table[key].sensorName, sensor_data_folder, timestamps_file_name, ctable.table[key].tfm, keyData, interval,1);

        // get extensions
        keyprefix=folderpath+"/"+ctable.table[key].folderName+"/"+ctable.table[key].folderName+"-";
        keysuffix=ctable.table[key].params["extension"];

        cout << keyData.size() << " timestamps read from " << sensor_data_folder+timestamps_file_name << endl;
        cout << ctable.table[key].folderName << " " << ctable.table[key].suffix << endl;

    }


    // initialise trajectory
    if(strcmp(key.c_str(),"r-gps")==0)
    {
        initialiseTraj(key,trajGPS,keyData);
    }
    else if(strcmp(key.c_str(),"l-gps")==0)
    {
        initialiseTraj(key,trajLowCostGPS,keyData);
    }
    else if(strcmp(key.c_str(),"cbus-poses")==0)
    {
        initialiseTraj(key,trajOdo,keyData);
    }

    long long int timeStart=20000000000000000llu;
    long long int timeEnd=0;

    singleSensor_=true;

    if(keyData.size()>0)
    {
       timeStart=keyData[0].timestamp;
       timeEnd=keyData[keyData.size()-1].timestamp;
    }
        


    cout <<"Time from: " <<timeStart << " to " << timeEnd  << " => " << (timeEnd-timeStart)/1000000 << " s" <<endl;

    imagecombin.setDateTime0(timeStart);

    currentkeyindex=-1;

    /*
                CONNECTION OF SIGNAL AND SLOTS AND THREAD elements
     */

    // JOB for f_l_cam
    connect(this,SIGNAL(terminateThreads()),&threadCamera_f_l_cam,SLOT(quit()));
    connect(this,SIGNAL(loadImage_f_l_cam(QString)),&loader_f_l_cam,SLOT(loadImage(QString)));
    loader_f_l_cam.setHeight(aimheight);
    connect(&loader_f_l_cam,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_f_l_cam(QImage)));
    loader_f_l_cam.moveToThread(&threadCamera_f_l_cam);
    threadCamera_f_l_cam.start();

    // JOB for f_r_cam
    connect(this,SIGNAL(terminateThreads()),&threadCamera_f_r_cam,SLOT(quit()));
    connect(this,SIGNAL(loadImage_f_r_cam(QString)),&loader_f_r_cam,SLOT(loadImage(QString)));
    loader_f_r_cam.setHeight(aimheight);
    connect(&loader_f_r_cam,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_f_r_cam(QImage)));
    loader_f_r_cam.moveToThread(&threadCamera_f_r_cam);
    threadCamera_f_r_cam.start();

    // JOB for f_m_cam
    connect(this,SIGNAL(terminateThreads()),&threadCamera_f_m_cam,SLOT(quit()));
    connect(this,SIGNAL(loadImage_f_m_cam(QString)),&loader_f_m_cam,SLOT(loadImage(QString)));
    loader_f_m_cam.setHeight(aimheight);
    connect(&loader_f_m_cam,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_f_m_cam(QImage)));
    loader_f_m_cam.moveToThread(&threadCamera_f_m_cam);
    threadCamera_f_m_cam.start();

    // JOB for b_cam
    connect(this,SIGNAL(terminateThreads()),&threadCamera_b_cam,SLOT(quit()));
    connect(this,SIGNAL(loadImage_b_cam(QString)),&loader_b_cam,SLOT(loadImage(QString)));
    loader_b_cam.setHeight(aimheight);
    connect(&loader_b_cam,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_b_cam(QImage)));
    loader_b_cam.moveToThread(&threadCamera_b_cam);
    threadCamera_b_cam.start();

    // JOB for cata_cam
    connect(this,SIGNAL(terminateThreads()),&threadCamera_cata_cam,SLOT(quit()));
    connect(this,SIGNAL(loadImage_cata_cam(QString)),&loader_cata_cam,SLOT(loadImage(QString)));
    loader_cata_cam.setHeight(aimheight);
    connect(&loader_cata_cam,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_cata_cam(QImage)));
    loader_cata_cam.moveToThread(&threadCamera_cata_cam);
    threadCamera_cata_cam.start();

    // JOB for fe_cam
    connect(this,SIGNAL(terminateThreads()),&threadCamera_fe_cam,SLOT(quit()));
    connect(this,SIGNAL(loadImage_fe_cam(QString)),&loader_fe_cam,SLOT(loadImage(QString)));
    loader_fe_cam.setHeight(aimheight);
    connect(&loader_fe_cam,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_fe_cam(QImage)));
    loader_fe_cam.moveToThread(&threadCamera_fe_cam);
    threadCamera_fe_cam.start();

    // JOB for webcam
    connect(this,SIGNAL(terminateThreads()),&threadCamera_webcam,SLOT(quit()));
    connect(this,SIGNAL(loadImage_webcam(QString)),&loader_webcam,SLOT(loadImage(QString)));
    loader_webcam.setHeight(aimheight);
    connect(&loader_webcam,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_webcam(QImage)));
    loader_webcam.moveToThread(&threadCamera_webcam);
    threadCamera_webcam.start();

    // JOB for RTK GPS
    trajGPS.setWidthHeight(aimwidth,aimheight);
    connect(this,SIGNAL(addNodeGPS(double, double,QColor)),&trajGPS,SLOT(addNode(double, double,QColor)));
    connect(&trajGPS,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_r_gps(QImage)));

    // JOB for low cost GPS
    trajLowCostGPS.setWidthHeight(aimwidth,aimheight);
    connect(this,SIGNAL(addNodeLowCostGPS(double, double,QColor)),&trajLowCostGPS,SLOT(addNode(double, double,QColor)));
    connect(&trajLowCostGPS,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_l_gps(QImage)));

    // JOB for the odo
    trajOdo.setWidthHeight(aimwidth,aimheight);
    connect(this,SIGNAL(addNodeOdo(double, double,QColor)),&trajOdo,SLOT(addNode(double, double,QColor)));
    connect(&trajOdo,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_cbus_poses(QImage)));


    // JOB for the horizontal lidar
    loaderLidHz.setWidthHeight(aimwidth,aimheight);
    loaderLidHz.setParameters(1.,maxLaserDist);
    connect(this,SIGNAL(loadImage_hz_laser(QString)),&loaderLidHz,SLOT(loadImage(QString)));
    connect(&loaderLidHz,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_hz_laser(QImage)));

    // JOB for the inclined lidar
    loaderLidIncl.setWidthHeight(aimwidth,aimheight);
    loaderLidIncl.setParameters(1.,maxLaserDist);
    connect(this,SIGNAL(loadImage_incl_laser(QString)),&loaderLidIncl,SLOT(loadImage(QString)));
    connect(&loaderLidIncl,SIGNAL(displayImage(QImage)),&imagecombin,SLOT(addImage_incl_laser(QImage)));

    // Job for time display
    connect(this,SIGNAL(addTime(long long int)),&imagecombin,SLOT(addTime(long long int)));

    // JOB for imagecombin
    connect(this,SIGNAL(terminateThreads()),&threadImagecombin,SLOT(quit()));
    connect(&imagecombin,SIGNAL(displayImage(QImage)),&ui->widget_,SLOT(drawPixmap(QImage)));


    connect(this,SIGNAL(addAdditionnalImage(QImage,bool)),&imagecombin,SLOT(addAdditionnalImage(QImage,bool)));




    imagecombin.moveToThread(&threadImagecombin);
    threadImagecombin.start();


    connect(ui->btnstart,SIGNAL(clicked()),this,SLOT(startUpdate()));

}

void MainWindow::startUpdate()
{   


    bool pause;
    mutPause_.lock();
    pause=pause_;
    mutPause_.unlock();

    //cout << "startUpdate() ; pause= "<<  pause<<  "  / fps= "<<fps <<endl;

    if(!stillstarted_)
    {
        stillstarted_=true;
        connect(&imagecombin,SIGNAL(displayImage(QImage)),this,SLOT(updateData(QImage)));
    }


    if(pause)
    {
        mutPause_.lock();
        pause_=false;
        mutPause_.unlock();

        ui->btnstart->setText("STOP REPLAY");

        if(fps == -1)
        {

            updateData();
        }
        else
        {
            // start reading
            QTimer *timer = new QTimer(this);
            connect(timer, SIGNAL(timeout()), this, SLOT(updateData()));
            int mytime=floor(1000./((double)fps));

            timer->start(mytime);
        }

    }
    else
    {
        mutPause_.lock();
        pause_=true;
        mutPause_.unlock();

        ui->btnstart->setText("START REPLAY");
    }
}


void MainWindow::updateData()
{
    bool pause;
    mutPause_.lock();
    pause=pause_;
    mutPause_.unlock();

    QString filename;
    QVector<string> vecRequiredImages;
    long int times;
    bool data_available=false;

    if(!pause)
    {





        if(strcmp(key.c_str(),"time")!=0)
        {
            currentkeyindex+=step;
            vecRequiredImages.append(key);
        }
        else
            currentkeyindex++;


        if(currentkeyindex<keyData.size())
        {
            if(typetimedisplay_ != 0)
                vecRequiredImages.append("time");



                data_available=true;
                imagecombin.setRequiredImages(vecRequiredImages);
                times=floor((double)keyData[currentkeyindex].timestamp/1000000.);
                ui->le_time->setText(QString("%1 s (UTC time)").arg(times));

                if(typetimedisplay_ != 0)
                    emit addTime(keyData[currentkeyindex].timestamp);


            if(data_available)
            {
                Unit localKeyData;
                localKeyData=keyData[currentkeyindex];


                // update data for  the key
                if(strcmp(key.c_str(),"l-gps")==0 || strcmp(key.c_str(),"r-gps")==0)
                {
                    double posX,posY;
                    gps_data data(localKeyData.reading);
                    posX=data.lat_l2e;
                    posY=data.lng_l2e;

                    if(strcmp(key.c_str(),"l-gps")==0)
                        emit addNodeLowCostGPS(posX,posY,Qt::blue);
                    else
                        emit addNodeGPS(posX,posY,Qt::blue);
                }
                else if(strcmp(key.c_str(),"cbus-poses")==0)
                {
                    double posX,posY;
                    pose_data data(localKeyData.reading);
                    posX=data.x;
                    posY=data.y;

                    emit addNodeOdo(posX,posY,Qt::blue);
                }
                else if(strcmp(key.c_str(),"hz-laser")==0 || strcmp(key.c_str(),"incl-laser")==0)
                {
                    getLidarFilename(localKeyData,keyprefix,filename);

                    if(strcmp(key.c_str(),"hz-laser")==0)
                        emit loadImage_hz_laser(filename);
                    else
                        emit loadImage_incl_laser(filename);
                }
                else if(strcmp(key.c_str(),"f-m-cam")==0 || strcmp(key.c_str(),"f-r-cam")==0 || strcmp(key.c_str(),"fe-cam")==0 || strcmp(key.c_str(),"f-l-cam")==0 || strcmp(key.c_str(),"b-cam")==0 || strcmp(key.c_str(),"cata-cam")==0 || strcmp(key.c_str(),"webcam")==0)
                {
                    // images
                    getImageFilename(localKeyData,keyprefix,keysuffix,filename);

                    loadImage(filename,key);
                }

            }
        }


    }
}

void MainWindow::updateData(QImage im)
{
    if(currentkeyindex<keyData.size())
    {
        currentImage_=im;
        currentImageTime_=keyData[currentkeyindex].timestamp;

        if(fps==-1)
        {
            updateData();
        }
    }
}

void MainWindow::getImageFilename(Unit & unit,std::string prefix,std::string suffix,QString & filename)
{
    ostringstream oss;
    oss.fill('0');

    oss << prefix << setw(10)<< unit.index+1  << suffix;


    filename=QString::fromStdString(oss.str());

}

void MainWindow::getLidarFilename(Unit & unit,std::string prefix,QString & filename)
{
    ostringstream oss;

    oss << prefix << unit.index+1 << "-0.txt";;


    filename=QString::fromStdString(oss.str());

}

MainWindow::~MainWindow()
{
    quit();
    delete ui;
}

void MainWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void MainWindow::keyPressEvent(QKeyEvent* event)
{
    if(event->key()==Qt::Key_Escape)
    {
        quit();

        // close the window
        close();

    }
    else if(event->key()==Qt::Key_Space)
    {
        startUpdate();
    }
    else if(event->key()==Qt::Key_S)
    {

        QString imageName=QString("Images/%1.png").arg(currentImageTime_);
        currentImage_.save(imageName,"PNG");

        cout <<"... Image " << imageName.toStdString() << " is saved"<< endl;
        ofstream myfile;
        myfile.open ("Images/Images.dates",ios::app);
        myfile << currentImageTime_ <<  " " <<  imageName.toStdString() <<endl;
        myfile.close();

    }
}

void MainWindow::quit()
{
    emit terminateThreads();

    threadCamera_f_l_cam.wait(200);
    threadCamera_f_l_cam.quit();

    threadCamera_f_r_cam.wait(200);
    threadCamera_f_r_cam.quit();

    threadCamera_f_m_cam.wait(200);
    threadCamera_f_m_cam.quit();

    threadCamera_b_cam.wait(200);
    threadCamera_b_cam.quit();

    threadCamera_cata_cam.wait(200);
    threadCamera_cata_cam.quit();

    threadCamera_fe_cam.wait(200);
    threadCamera_fe_cam.quit();

    threadCamera_webcam.wait(200);
    threadCamera_webcam.quit();

    threadImagecombin.wait(200);
    threadImagecombin.quit();
}

void MainWindow::loadImage(QString filename,string ent)
{


    if(strcmp(ent.c_str(),"f-l-cam")==0)
    {
        //        cout << "loadImage1: "<< filename.toStdString() <<" / " << ent <<endl;
        emit loadImage_f_l_cam(filename);
    }
    else if(strcmp(ent.c_str(),"f-r-cam")==0)
    {
        //        cout << "loadImage_f_r_cam: "<< filename.toStdString() <<" / " << ent <<endl;
        emit loadImage_f_r_cam(filename);
    }
    else if(strcmp(ent.c_str(),"f-m-cam")==0)
    {
        //        cout << "loadImage_f_m_cam: "<< filename.toStdString() <<" / " << ent <<endl;
        emit loadImage_f_m_cam(filename);
    }
    else if(strcmp(ent.c_str(),"b-cam")==0)
    {
        //        cout << "loadImage_b_cam: "<< filename.toStdString() <<" / " << ent <<endl;
        emit loadImage_b_cam(filename);
    }
    else if(strcmp(ent.c_str(),"cata-cam")==0)
    {
        //        cout << "loadImage_cata_cam: "<< filename.toStdString() <<" / " << ent <<endl;
        emit loadImage_cata_cam(filename);
    }
    else if(strcmp(ent.c_str(),"fe-cam")==0)
    {
        //        cout << "loadImage_fe_cam: "<< filename.toStdString() <<" / " << ent <<endl;
        emit loadImage_fe_cam(filename);
    }
    else if(strcmp(ent.c_str(),"webcam")==0)
    {
        //        cout << "loadImage_webcam: "<< filename.toStdString() <<" / " << ent <<endl;
        emit loadImage_webcam(filename);
    }

}

void MainWindow::initialiseTraj(std::string entity,TrajectoryToImage& traj,std::vector<Unit> & data)
{
    // check for window size
    // and peuplate  vecPosT vector
    double minX=10000000;
    double minY=10000000;
    double maxX=-10000000;
    double maxY=-10000000;

    double posX,posY;

    for(unsigned int i=0;i<data.size();i++)
    {
        if(strcmp(entity.c_str(),"l-gps")==0 || strcmp(entity.c_str(),"r-gps")==0)
        {
            gps_data donnees(data[i].reading);
            posX=donnees.lat_l2e;
            posY=donnees.lng_l2e;
        }
        else // odo
        {
            pose_data donnees(data[i].reading);
            posX=donnees.x;
            posY=donnees.y;
        }

        if(minX>posX)
            minX=posX;
        else if(maxX<posX)
            maxX=posX;

        if(minY>posY)
            minY=posY;
        else if(maxY<posY)
            maxY=posY;


    }


    minX-=5.;
    maxX+=5.;

    minY-=5.;
    maxY+=5.;

    cout << "Data: ( " << setprecision(10)<< minX<< " , "<< minY << " ) < (x,y) < "<< "( " << maxX<< " , "<< maxY << " ) "<< endl;

    traj.setSize(minX,minY,maxX-minX,maxY-minY);

}

void MainWindow::loadSatelliteImage(QString filename,int height,int width)
{
    QImage imSat=QImage(filename);
    if(!imSat.isNull())
    {
        imSat=imSat.scaledToHeight(height);
        imSat=imSat.mirrored(false,true);

        QPixmap fullPixmap=QPixmap(width,height);
        fullPixmap.fill(Qt::white);

        QImage fullImage=fullPixmap.toImage ();


        QRectF source=QRectF(0,0,imSat.width(),imSat.height());
        QRectF target=QRectF((fullImage.width()-imSat.width())/2,(fullImage.height()-imSat.height())/2,imSat.width(),imSat.height());

        QPainter paint;
        paint.begin(&fullImage);
        paint.setRenderHints(QPainter::Antialiasing);
        paint.drawImage(target,imSat,source);
        paint.end();

        imSat=fullImage;

        imagecombin.addAdditionnalImage(imSat,true);

    }
}


