/****************************************************************************
 *
 * $Id:
 *
 * Copyright (C) 1998-2006 Lasmea. All rights reserved.
 *
 * This software was developed at:
 * Institut Pascal- UMR 6602
 * 63 175 Aubière Cedex
 * http://ip.univ-bpclermont.fr
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE included in the packaging of this file.
 *
 * Licensees holding valid IPDS Professional Edition licenses may
 * use this file in accordance with the IPDS Commercial License
 * Agreement provided with the Software.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Contact . if any conditions of this licensing are
 * not clear to you.
 *
 *
 * Author: Jonathan Courbon - Hemanth Korrapati -
 *
 *****************************************************************************/


#include <iostream>
#include <sstream>
#include <vector>
#include <string>

#include "config.h"

#if QTDETECTEDVERSION == 4
    #include <QtGui/QApplication>
#else
    #include <QtWidgets/QApplication>
#endif



#include "containers.h"
#include "config.h"

#include "imagecombination.h"

#include "mainwindow.h"



void usage()
{
    std::cout << "============================" << std::endl;
    std::cout << "  Viewer Single Sensor" << std::endl;
    std::cout << "============================" << std::endl;
    std::cout << "Usage: ./ViewerSingleSensor sequencePath sensorAlias [displaysensor] [displaytime]" << std::endl;
    std::cout << "   The sensorAlias must be one of the following value: f-l-cam ,f-r-cam, b-cam, f-m-cam, fe-cam, webcam, cata-cam, r-gps, l-gps, cbus-poses, hz-laser,incl-laser" << std::endl;
    std::cout << "   displaysensor: if it is equal to true (default), the name of the sensor is added in the image" << std::endl;
    std::cout << "   displaytime: if it is equal to 0: the time is not displayed, 1 (default): the UTC time is added in the image, 2: the time expressed in minutes and seconds from the begiing of the acuisition is added to the image" << std::endl;
}



int main(int argc, char* argv[])
{

    QApplication a(argc, argv);
    setlocale (LC_ALL,"C");

    
    if(argc>=3)
    {
      std::string sequencePath(argv[1]);
      std::string sensorAlias(argv[2]);
      
      // by default, the sensor name is added
      bool addSensorsName=true;
      // by default, the UTC time is displayed 
      int typetimedisplay=1;
      
      if(argc>=4)
      {
	addSensorsName=atoi(argv[4]);
	if(argc>=5)
	{
	  typetimedisplay=atoi(argv[5]);
	}
      }
      
      bool valid=false;
      
      if(strcmp(sensorAlias.c_str(),"f-l-cam")==0
            || strcmp(sensorAlias.c_str(),"f-r-cam")==0
            || strcmp(sensorAlias.c_str(),"b-cam")==0
            || strcmp(sensorAlias.c_str(),"f-m-cam")==0
            || strcmp(sensorAlias.c_str(),"fe-cam")==0
            || strcmp(sensorAlias.c_str(),"webcam")==0
            || strcmp(sensorAlias.c_str(),"cata-cam")==0
            || strcmp(sensorAlias.c_str(),"r-gps")==0
            || strcmp(sensorAlias.c_str(),"l-gps")==0
            || strcmp(sensorAlias.c_str(),"cbus-poses")==0
            || strcmp(sensorAlias.c_str(),"hz-laser")==0
            || strcmp(sensorAlias.c_str(),"incl-laser")==0)
      {
	
	
	ImageCombination::MODE mode;
	int step;
	std::string key;
	
	// image created by ImageCombination object will  be of size 640x480
	int aimwidth=640;
	int aimheight=480;
	
	// data will be displayed as fast as possible
	int fps=-1;

	// maximum range to display range sensor impacts
	double maxLaserDist=50;
	
	long long int startingtime=-1;
        long long int endingtime=-1;
    
    
	mode=ImageCombination::SINGLE;
	
	if(strcmp(sensorAlias.c_str(),"l-gps")==0)
        {
	    std::cout << "GPS trajectory will be displayed with all data" << std::endl; 
            step=1;
            aimwidth=800;
            aimheight=600;
	    valid=true;
        }
        else if(strcmp(sensorAlias.c_str(),"r-gps")==0)
        {
	    std::cout << "GPS trajectory will be displayed with one data over 20" << std::endl; 
            step=20;
            aimwidth=800;
            aimheight=600;
	    valid=true;
        }
        else if(strcmp(sensorAlias.c_str(),"hz-laser")==0 || strcmp(sensorAlias.c_str(),"incl-laser")==0)
        {
	    std::cout << "Range sensor data will be displayed with one data over 10" << std::endl; 
            step=10;
	    valid=true;
        }
        else if(strcmp(sensorAlias.c_str(),"cbus-poses")==0)
        {
	    std::cout << "2D dead-recockned trajectory will be displayed with one data over 10" << std::endl; 
            step=10;
            aimwidth=800;
            aimheight=600;
	    valid=true;
        }
        else if(strcmp(sensorAlias.c_str(),"webcam")==0)
        {
	    std::cout << "Images from the webcam will be displayed one over 5" << std::endl; 
            step=5;
	    valid=true;
        }
        else
        {
            step=1;
            fps=-1;
	    valid=true;


        }


        
	
	if(valid)
        {

	  MainWindow window(sequencePath,sensorAlias,aimwidth,aimheight,fps,typetimedisplay,mode,step,addSensorsName,maxLaserDist,startingtime,endingtime);
	  window.show();
	  return a.exec();
        }
	
      }
      else
      {
	  
	usage();
	
      }
      
      

    
    
    }
    else
    {
      usage();
      return 0;
    }


 


}





